// Copyright (C) 2019  Joseph Artsimovich <joseph.artsimovich@gmail.com>, 4lex4 <4lex49@zoho.com>
// Use of this source code is governed by the GNU GPLv3 license that can be found in the LICENSE file.

#ifndef SCANTAILOR_IMAGEPROC_BYTEORDER_H_
#define SCANTAILOR_IMAGEPROC_BYTEORDER_H_

#ifdef _WIN32

#include <winsock2.h>

#else
#include <netinet/in.h>
#endif

#endif
